#include <stdint.h>       // Find types like "uint8_t"
#include <stdbool.h>      // Find type "bool"
#include <SPI.h>          // Arduino SPI library
#include <SD.h>           // Arduino SD card library

#include "Eve2_81x.h"     // Matrix Orbital Eve2 Library      
#include "Display_Timings.h" // TFT timings
#include "Arduino_AL.h"   // Hardware abstraction layer for Arduino
#include "process.h"      // More application level code


// The file of type File and named myFile is global and used for any and all files.  
// Your job is to make sure you are not trying to open it twice.
File myFile;

// A global buffer for use with debug messages.  It is always used up immediately so 
// never retains initialized values.  Prevent further allocations of RAM for logging
char LogBuf[LogBufSize];

void setup()
{
  // Initializations.  Order is important
  GlobalInit();
  FT81x_Init();
  SD_Init();

  Calibrate_Cap_Touch();  
  if (!LoadTouchMatrix())
  {
    // We failed to read calibration matrix values from the SD card.       
    MakeScreen_Calibrate();
    
    if( SaveTouchMatrix() )  // If it saved, then
      LoadTouchMatrix();     // read it back from the file to compare values   
  }  
}

void loop()
{    
  Waves(); 
}

static uint8_t rate = 2;
static int16_t x,y,tx;
static uint8_t beats[10];
static char beats_Incr[10] = {-10,10,5,-5,-20,20,12,-12,-5,5};
static uint16_t add2write = 0;

/* Optimized implementation of sin and cos table - precision is 16 bit */
uint16_t sintab[] = {
    0, 402, 804, 1206, 1607, 2009, 2410, 2811, 3211, 3611, 4011, 4409, 4807, 5205, 5601, 5997, 6392,
    6786, 7179, 7571, 7961, 8351, 8739, 9126, 9511, 9895, 10278, 10659, 11038, 11416, 11792, 12166, 12539,
    12909, 13278, 13645, 14009, 14372, 14732, 15090, 15446, 15799, 16150, 16499, 16845, 17189, 17530, 17868,
    18204, 18537, 18867, 19194, 19519, 19840, 20159, 20474, 20787, 21096, 21402, 21705, 22004, 22301, 22594,
    22883, 23169, 23452, 23731, 24006, 24278, 24546, 24811, 25072, 25329, 25582, 25831, 26077, 26318, 26556, 26789,
    27019, 27244, 27466, 27683, 27896, 28105, 28309, 28510, 28706, 28897, 29085, 29268, 29446, 29621, 29790, 29955,
    30116, 30272, 30424, 30571, 30713, 30851, 30984, 31113, 31236, 31356, 31470, 31580, 31684, 31785, 31880, 31970,
    32056, 32137, 32213, 32284, 32350, 32412, 32468, 32520, 32567, 32609, 32646, 32678, 32705, 32727, 32744, 32757,
    32764, 32767, 32764};


int16_t qsin(uint16_t a)
{
    uint8_t f;
    int16_t s0,s1;

    if (a & 32768)
    {
      return -qsin(a & 32767);
    }
    if (a & 16384)
    {
      a = 32768 - a;
      f = a & 127;
      s0 = pgm_read_word(sintab + (a >> 7));
      s1 = pgm_read_word(sintab + (a >> 7) + 1);
      return (s0 + ((int32_t)f * (s1 - s0) >> 7));
    }
}

/* cos funtion */
int16_t qcos(uint16_t a)
{
    return (qsin(a + 16384));
}

static bool istouch()
{
    return !(rd16(REG_CTOUCH_TOUCH_XY + RAM_REG) & 0x8000);
}

void Sine_wave(uint8_t amp)
{
    static uint8_t played = 0,change=0;
    x+=rate;
    if(x>HSIZE)
    {
      x  = 0;
    }
    y = (VSIZE/2) + ((int32_t)amp*qsin(-65536*x/(25*rate))/65536);
    if(played==0 &&  change < y)
    { 
      played = 1;
      wr16(REG_SOUND + RAM_REG,0x50);
      wr8(REG_VOL_SOUND + RAM_REG,100);      
      wr8(REG_PLAY + RAM_REG,1); 
    }
    if(change > y)
    {
      played = 0;
    }
    change = y;
    wr16(RAM_G+(x/rate)*4,VERTEX2F(x*16,y*16));  
}

void Sawtooth_wave(uint8_t amp)
{
    static uint16_t temp=0;
    static uint8_t pk = 0;
    x+=rate;
    if(x>HSIZE)
    {
      x  = 0;
    }
    temp+=2; 
    if(temp>65535L)
    {
      temp = 0;
    }
    y = (temp % amp);
    pk = y/(amp-2);
    if(pk) 
    {
      wr16(REG_SOUND + RAM_REG,0x50);
      wr8(REG_VOL_SOUND + RAM_REG,100);      
      wr8(REG_PLAY + RAM_REG,1);
    }
    y = (VSIZE/2)-y;
    wr16(RAM_G+(x/rate)*4,VERTEX2F(x*16,y*16)); 
}

void Triangle_wave(uint8_t amp)
{
    static uint16_t temp=0;
    static uint8_t pk = 0,dc=0,p=0;
    x+=rate;
    if(x>HSIZE)
    { 
      x  = 0;
    }
    temp+=2; 
    if(temp>65535L)
    {
      temp = 0;
    }
    y = (temp % amp);
    pk = (y/(amp-2))%2;
    dc = (temp / amp)%2;  
    if(pk) 
    { 
      if(p==0)
      {
        p=1;         
        wr16(REG_SOUND + RAM_REG,0x50);
        wr8(REG_VOL_SOUND + RAM_REG,100);      
        wr8(REG_PLAY + RAM_REG,1);
      } 
      else  
      {
        p=0;
      }
    }
    if(dc)
    {
      y = (VSIZE/2) -(amp-y);  
    }
    else
    {
      y = (VSIZE/2) - y;
    }        
    wr16(RAM_G+(x/rate)*4,VERTEX2F(x*16,y*16));  
}

static uint16_t temp_x,temp_p,temp_y,en;

void Heartbeat()
{
  uint16_t tval;
  y = VSIZE/2;
  for(tval=0;tval<10;tval++)
  {
    y = y+(beats_Incr[tval]*5);
    beats[tval] = y;
    //printf(" \n tval=%d   beats=   %d",tval,beats[tval]);
  }
  x+=rate;
  if(x>HSIZE)
  { 
    x  = 0;temp_p = 0;
    temp_y=0;  
    y=VSIZE/2;
    en=0;
    temp_x=0;
  }
  tx = 5*rate;
  tx = ((temp_p+1)*tx) + temp_p*temp_x;
  if(tx<=x)
  { 
    if(0==en)    
    en = 1;
  }
  if(en==1)
  {
    if(y!=beats[temp_y])
    {
        y += beats_Incr[temp_y] * 5;
        temp_y++;
        if(y==(VSIZE/2)+beats_Incr[4] * 5)
        {
          wr16(REG_SOUND + RAM_REG,0x50);
          wr8(REG_VOL_SOUND + RAM_REG,100);      
          wr8(REG_PLAY + RAM_REG,1); 
        }
    }
    else
    {
        temp_y++;
        if(temp_y>9) 
        {
          temp_y = 0;   
          temp_p++;
          en = 0;  
          temp_x = x - tx;
        }
    }
  }
  wr32(RAM_G+(x/rate)*4,VERTEX2F(x*16,y*16));
}

void Gpu_Radiobutton(int16_t x,int16_t y,uint32_t bgcolor,uint32_t fgcolor,uint8_t psize,uint8_t tag,uint8_t option)
{
    uint8_t check_size = psize/2;
    Send_CMD(SAVE_CONTEXT());
    Send_CMD(COLOR_RGB((bgcolor >> 16 & 0xff),(bgcolor >> 8 & 0xff),(bgcolor & 0xff)));
    Send_CMD(TAG(tag));
    Send_CMD(POINT_SIZE(psize*16));
    Send_CMD(BEGIN(POINTS));
    Send_CMD(VERTEX2F(x*16,y*16));
    if(tag == option)
    {
        Send_CMD(COLOR_RGB((fgcolor >> 16 & 0xff),(fgcolor >> 8 & 0xff),(fgcolor & 0xff)));
        Send_CMD(POINT_SIZE(check_size*16));
        Send_CMD(VERTEX2F(x*16,y*16));
    }
    Send_CMD(RESTORE_CONTEXT());  
}

void Waves()
{
    int16_t i,xx=0;
    uint16_t tval,th_to,amp;
    uint8_t  temp[480],hide_x=0,fg=1,tag,opt=3;
    // ========  Bg =========================================================
    for(tval=0;tval<=VSIZE/2;tval++)  
    {
        temp[VSIZE-tval] = temp[tval] = (tval*0.90);   
    }

    y = VSIZE/2;
    for(tval=0;tval<10;tval++)
    {
        y = y+(beats_Incr[tval]*5);
        beats[tval] = y;
    }

    for(tval=0;tval<HSIZE;tval+=rate)
    {
        wr32(RAM_G+((tval/rate)*4),VERTEX2F(tval*16,y*16));     
    }
    add2write = 0;

    while(1)
    {


        // ========  Menu  =========================================================
        if(istouch())
        {
          fg = 1;
        }
        if(fg)
        { 
          th_to=0; 
          if(hide_x>0)
          {
            hide_x=0; 
          }
          else
          {
            fg = 0;
          }
        }
        else 
        { 
          th_to++;
          if(th_to > 200) 
          {
            if( hide_x < 140)
            { 
              hide_x +=3;
            }   
            else
            {
              th_to = 0;
            }     
          }
        }
        //==========Option =========================================================    
        //tag = App_Read_Tag(phost);
        tag = rd8(REG_CTOUCH_TAG + RAM_REG);
        Log("%d\n", tag);
        Log("%d\n", rd8(REG_CTOUCH_TOUCH_XY + RAM_REG));
        Log("%d\n", rd8(REG_CTOUCH_TAG + RAM_REG));
        if(tag!=0)
        {
            x = 0;   temp_p = 0;en = 0; temp_x = 0; temp_y = 0; //reset
            if(tag>2)  {
              opt = tag;
            }
            if(tag==1)
            {
              if(rate>1)
              {
                rate--;
              }
            }
            if(tag==2)
            {
              if(rate<9)
              rate++;
            }     
            y = (VSIZE/2);
            for(tval=0;tval<HSIZE;tval+=rate)
            {
                wr32(RAM_G+((tval/rate)*4),VERTEX2F(tval*16,y*16));     
            }
            add2write = 0;
        }    
        //========= Signals ========================================================
        amp = 100;
        switch(opt)
        {
        case 5:

            Triangle_wave(amp);
            break;

        case 4:
            Sawtooth_wave(amp);
            break;

        case 3:
            Sine_wave(amp);
            break;

        case 6:
            amp = 50;
            Heartbeat();          
            break;
        }
        //=========Display list start===================================================    
        //Gpu_CoCmd_Dlstart(phost); 
        Send_CMD(CMD_DLSTART);
        Send_CMD(CLEAR(1,1,1));
        Send_CMD(COLOR_RGB(0x12,0x4A,0x26)); 
        Send_CMD(BEGIN(RECTS));
        Send_CMD(TAG(0));
        Send_CMD(VERTEX2F(0,0));
        Send_CMD(COLOR_RGB(0x1B,0xE0,0x67));  

        Send_CMD(COLOR_RGB(0x1B,0xE0,0x67));
        Send_CMD(LINE_WIDTH(2*16));
        Send_CMD(BEGIN(LINE_STRIP));
        Cmd_Append(RAM_G,(x/rate)*4);
        
        Send_CMD(END());

        Send_CMD(BEGIN(LINE_STRIP));  
        if((x/rate)<(HSIZE/rate)-(50/rate))            // else it screw up
        {
          Cmd_Append(RAM_G+(x/rate)*4+((50/rate)*4),((HSIZE/rate)*4)-((x/rate)*4)-((50/rate)*4));
        }

        Send_CMD(END());  

        Send_CMD(POINT_SIZE(6*16));
        Send_CMD(BEGIN(POINTS));
        Send_CMD(VERTEX2F(x*16,y*16));  
        Send_CMD(END());
        Send_CMD(COLOR_RGB(0xff,0xff,0xff));
        Send_CMD(COLOR_A(100)); 
        Send_CMD(BEGIN(EDGE_STRIP_R));
        Send_CMD(VERTEX2F((hide_x+HSIZE-140)*16,0));
        Send_CMD(VERTEX2F((hide_x+HSIZE-140)*16,VSIZE*16));
        Send_CMD(COLOR_A(255));             
        Gpu_Radiobutton(hide_x+HSIZE-120,VSIZE-30 ,0xffffff,0,20,5,opt);
        Gpu_Radiobutton(hide_x+HSIZE-120,VSIZE-75,0xffffff,0,20,4,opt);
        Gpu_Radiobutton(hide_x+HSIZE-120,VSIZE-120,0xffffff,0,20,3,opt);
        Gpu_Radiobutton(hide_x+HSIZE-120,VSIZE-165,0xffffff,0,20,6,opt);


        Cmd_Text(hide_x+HSIZE-90,VSIZE-30 ,23,OPT_CENTERY,"Triangle");        
        Cmd_Text(hide_x+HSIZE-90,VSIZE-75,23,OPT_CENTERY,"Sawtooth");
        Cmd_Text(hide_x+HSIZE-90,VSIZE-120,23,OPT_CENTERY,"Sine");        
        Cmd_Text(hide_x+HSIZE-90,VSIZE-165,23,OPT_CENTERY,"ECG");
        Cmd_Text((hide_x+HSIZE-120),20,30,OPT_CENTERY|OPT_CENTERX,"-");
        Cmd_Text((hide_x+HSIZE-20),20,30,OPT_CENTERY|OPT_CENTERX,"+");
        
        Cmd_Text((hide_x+HSIZE-95),15,22,0,"Rate:");
        Cmd_Number((hide_x+HSIZE-50),15,22,0,rate);   
        Cmd_Text((hide_x+HSIZE-120),80,23,0,"Pk:");
        Cmd_Number((hide_x+HSIZE-70),80,23,0,amp);    
        
        Send_CMD(COLOR_A(50));
        Send_CMD(POINT_SIZE(20*16));
        Send_CMD(BEGIN(POINTS));
        Send_CMD(TAG(1));
        Send_CMD(VERTEX2F((hide_x+HSIZE-120)*16,20*16));   
        Send_CMD(TAG(2));
        Send_CMD(VERTEX2F((hide_x+HSIZE-20)*16,20*16)); 

        Send_CMD(DISPLAY());
        Send_CMD(CMD_SWAP);  
        UpdateFIFO();  
        //==========================End==================================================    
    }
}

// ************************************************************************************
// Following are wrapper functions for C++ Arduino functions so that they may be      *
// called from outside of C++ files.  These are also your opportunity to use a common *
// name for your hardware functions - no matter the hardware.  In Arduino-world you   *
// interact with hardware using Arduino built-in functions which are all C++ and so   *
// your "abstraction layer" must live in this xxx.ino file where C++ works.           *
//                                                                                    *
// Incidentally, this is also an alternative to ifdef-elif hell.  A different target  *
// processor or compiler will include different files for hardware abstraction, but   *
// the core "library" files remain unaltered - and clean.  Applications built on top  *
// of the libraries need not know which processor or compiler they are running /      *
// compiling on (in general and within reason)                                        *
// ************************************************************************************

void GlobalInit(void)
{
  Serial.begin(115200);
  while (!Serial) {;}                    // wait for serial port to connect.
  
  // Keys Initialization
  Init_Keys();

  // Matrix Orbital Eve display interface initialization
  pinMode(EvePDN_PIN, OUTPUT);            // Pin setup as output for Eve PDN pin.
  digitalWrite(EvePDN_PIN, LOW);          // Apply a resetish condition on Eve
  pinMode(EveChipSelect_PIN, OUTPUT);     // SPI CS Initialization
  digitalWrite(EveChipSelect_PIN, HIGH);  // Deselect Eve
  pinMode(EveAudioEnable_PIN, OUTPUT);    // Audio Enable PIN
  digitalWrite(EveAudioEnable_PIN, LOW);  // Disable Audio
  digitalWrite(EvePDN_PIN, HIGH);          // Apply a resetish condition on Eve
  SPI.begin();                            // Enable SPI
  
}

// Send a single byte through SPI
void SPI_WriteByte(uint8_t data)
{
  SPI.beginTransaction(SPISettings(100000, MSBFIRST, SPI_MODE0));
  digitalWrite(EveChipSelect_PIN, LOW);

  SPI.transfer(data);
      
  digitalWrite(EveChipSelect_PIN, HIGH);
  SPI.endTransaction();
}

// Send a series of bytes (contents of a buffer) through SPI
void SPI_WriteBuffer(uint8_t *Buffer, uint32_t Length)
{
  SPI.beginTransaction(SPISettings(100000, MSBFIRST, SPI_MODE0));
  digitalWrite(EveChipSelect_PIN, LOW);

  SPI.transfer(Buffer, Length);
      
  digitalWrite(EveChipSelect_PIN, HIGH);
  SPI.endTransaction();
}

// Send a byte through SPI as part of a larger transmission.  Does not enable/disable SPI CS
void SPI_Write(uint8_t data)
{
//  Log("W-0x%02x\n", data);
  SPI.transfer(data);
}

// Read a series of bytes from SPI and store them in a buffer
void SPI_ReadBuffer(uint8_t *Buffer, uint32_t Length)
{
  uint8_t a = SPI.transfer(0x00); // dummy read

  while (Length--)
  {
    *(Buffer++) = SPI.transfer(0x00);
  }
}

// Enable SPI by activating chip select line
void SPI_Enable(void)
{
  SPI.beginTransaction(SPISettings(100000, MSBFIRST, SPI_MODE0));
  digitalWrite(EveChipSelect_PIN, LOW);
}

// Disable SPI by deasserting the chip select line
void SPI_Disable(void)
{
  digitalWrite(EveChipSelect_PIN, HIGH);
  SPI.endTransaction();
}

void Init_Keys(void)
{
  // Pin setup as input for each button pin
  pinMode(Button1_PIN, INPUT);
  pinMode(Button2_PIN, INPUT);
  pinMode(Button3_PIN, INPUT);
}

void Eve_Reset_HW(void)
{
  // Reset Eve
  SetPin(EvePDN_PIN, 0);                    // Set the Eve PDN pin low
  MyDelay(50);                             // delay
  SetPin(EvePDN_PIN, 1);                    // Set the Eve PDN pin high
  MyDelay(100);                            // delay
}

void DebugPrint(char *str)
{
  Serial.print(str);
}

// A millisecond delay wrapper for the Arduino function
void MyDelay(uint32_t DLY)
{
  uint32_t wait;
  wait = millis() + DLY; while(millis() < wait);
}

// An abstracted pin write that may be called from outside this file.
void SetPin(uint8_t pin, bool state)
{
  digitalWrite(pin, state); 
}

// An abstracted pin read that may be called from outside this file.
uint8_t ReadPin(uint8_t pin)
{
  return(digitalRead(pin));
}

// Read the touch digitizer calibration matrix values from the Eve and write them to a file
bool SaveTouchMatrix(void)
{
  uint8_t count = 0;
  uint32_t data;
  uint32_t address = REG_TOUCH_TRANSFORM_A + RAM_REG;
  
//  Log("Enter SaveTouchMatrix\n");
  
  // If the file exists already from previous run, then delete it.
  if(SD.exists("tmatrix.txt"))
    SD.remove("tmatrix.txt");
  delay(50);
  
  FileOpen("tmatrix.txt", FILEWRITE);
  if(!myFileIsOpen())
  {
    Log("No create file\n");
    FileClose();
    return false;
  }
  
  do
  {
    data = rd32(address + (count * 4));
    Log("TMw: 0x%08lx\n", data);
    FileWriteByte(data & 0xff);                // Little endian file storage to match Eve
    FileWriteByte((data >> 8) & 0xff);
    FileWriteByte((data >> 16) & 0xff);
    FileWriteByte((data >> 24) & 0xff);
    count++;
  }while(count < 6);
  FileClose();
  Log("Matrix Saved\n");
}

// Read the touch digitizer calibration matrix values from a file and write them to the Eve.
bool LoadTouchMatrix(void)
{
  uint8_t count = 0;
  uint32_t data;
  uint32_t address = REG_TOUCH_TRANSFORM_A + RAM_REG;
  
  FileOpen("tmatrix.txt", FILEREAD);
  if(!myFileIsOpen())
  {
    Log("tmatrix.txt not open\n");
    FileClose();
    return false;
  }
  
  do
  {
    data = FileReadByte() +  ((uint32_t)FileReadByte() << 8) + ((uint32_t)FileReadByte() << 16) + ((uint32_t)FileReadByte() << 24);
    Log("TMr: 0x%08lx\n", data);
    wr32(address + (count * 4), data);
    count++;
  }while(count < 6);
  
  FileClose();
  Log("Matrix Loaded \n");
  return true;
}

void SD_Init(void)
{
//  Log("Initializing SD card...\n");
  if (!SD.begin(SDChipSelect_PIN)) 
  {
    Log("SD initialization failed!\n");
    return;
  }
  Log("SD initialization done\n");
}

// ************************************************************************************
// Following are abstracted file operations for Arduino.  This is possible by using a * 
// global pointer to a single file.  It is enough for our needs and it hides file     *
// handling details within the abstraction.                                           *
// ************************************************************************************
void FileOpen(char *filename, uint8_t mode)
{
  // Since one also loses access to defined values like FILE_READ from outside the .ino
  // I have been forced to make up values and pass them here (mode) where I can use the 
  // Arduino defines.
  switch(mode)
  {
  case FILEREAD:
    myFile = SD.open(filename, FILE_READ);
    break;
  case FILEWRITE:
    myFile = SD.open(filename, FILE_WRITE);
    break;
  default:;
  }
}

void FileClose(void)
{
  myFile.close();
  if(myFileIsOpen())
  {
    Log("Failed to close file\n");
  }
}

// Read a single byte from a file
uint8_t FileReadByte(void)
{
  return(myFile.read());
}

// Read bytes from a file into a provided buffer
void FileReadBuf(uint8_t *data, uint32_t NumBytes)
{
  myFile.read(data, NumBytes);
}

void FileWriteByte(uint8_t data)
{
  myFile.write(data);
}

uint32_t FileSize(void)
{
  return(myFile.size());
}

uint32_t FilePosition(void)
{
  return(myFile.position());
}

bool FileSeek(uint32_t offset)
{
  return(myFile.seek(offset));
}

bool myFileIsOpen(void)
{
  if(myFile)
    return true;
  else
    return false;
}
