// Process.c is the application layer.  All function calls are hardware ambivalent.

#include <stdint.h>              // Find integer types like "uint8_t"  
#include "Eve2_81x.h"            // Matrix Orbital Eve2 Driver
#include "Arduino_AL.h"          // include the hardware specific abstraction layer header for the specific hardware in use.
#include "process.h"             // Every c file has it's header and this is the one for this file
#include "Display_Timings.h" // TFT timings

// A calibration screen for the touch digitizer
void MakeScreen_Calibrate(void)
{
  Log("Enter Calibrate\n");
  
  Send_CMD(CMD_DLSTART);
  Send_CMD(CLEAR_COLOR_RGB(0,0,0));
  Send_CMD(CLEAR(1,1,1));
  Cmd_Text(100, 240, 27, OPT_CENTER, "Tap on the dots");
  Cmd_Calibrate(0);                                           // This widget generates a blocking screen that doesn't unblock until 3 dots have been touched
  Send_CMD(DISPLAY());
  Send_CMD(CMD_SWAP);
  UpdateFIFO();                                               // Trigger the CoProcessor to start processing commands out of the FIFO
  
  Wait4CoProFIFOEmpty();                                      // wait here until the coprocessor has read and executed every pending command.
  MyDelay(100);

  Log("Leaving Calibrate\n");
}

// This define is for the size of the buffer we are going to use for data transfers.  It is 
// sitting here so uncomfortably because it is a silly tiny buffer in Arduino Uno and you
// will want a bigger one if you can get it.  Redefine this and add a nice buffer to Load_ZLIB()
#define COPYBUFSIZE WorkBuffSz

void Calibrate_Cap_Touch(void)
{
  uint32_t Remaining;
  uint16_t ReadBlockSize = 0;
  uint32_t Add_CMD = RAM_CMD;
  uint8_t SD_data = 0;
  uint32_t Remaining_Counter = 0;

  // Open the file on SD card by name
  FileOpen("Cap_Cal.bin", FILEREAD);
  if(!myFileIsOpen())
  {
    Log("Cap_Cal.bin not open\n");
    FileClose();
    return false;
  }
  Log("Loaded\n");
  
  Remaining = FileSize();                                    // Store the size of the currently opened file
   while (Remaining)
  {
    if (Remaining > COPYBUFSIZE)
      ReadBlockSize = COPYBUFSIZE;
    else
      ReadBlockSize = Remaining;    
    Log("Block Begin\n");
    FileReadBuf(LogBuf, ReadBlockSize);                        // Read a block of data from the file
    //Add_CMD = WriteBlockRAM(Add_CMD, LogBuf, ReadBlockSize); 
    CoProWrCmdBuf(LogBuf, ReadBlockSize);
    Remaining -= ReadBlockSize;                                // Reduce remaining data value by amount just read
    Log("Block End\n");
    UpdateCMDFIFO();
    Log("Waiting End\n");
    Wait4CoProFIFOEmpty(); 
  }    
  FileClose();
  Log("Close File");

  wr8(RAM_REG + REG_CPU_RESET,2);
  wr16(RAM_REG + REG_CYA_TOUCH, 0x05D0);
  wr8(RAM_REG + REG_GPIOX_DIR, (rd8(RAM_REG + REG_GPIOX_DIR) | 0x08));
  wr8(RAM_REG + REG_GPIOX, (rd8(RAM_REG + REG_GPIOX) & 0xF7));
  MyDelay(1);
  wr8(RAM_REG + REG_CPU_RESET,0);
  MyDelay(100);
  wr8(RAM_REG + REG_GPIOX_DIR, (rd8(RAM_REG + REG_GPIOX_DIR) & 0xF7));
  MyDelay(100);
  wr8(RAM_REG + REG_CTOUCH_MODE,3);
  MyDelay(100);
  wr8(RAM_REG + REG_CTOUCH_EXTEND,1);  
  MyDelay(100);
  
}

// Is there a physical key press?
uint8_t CheckKeys(void)
{
  if(!ReadPin(Button1_PIN))
    return(1);
  if(!ReadPin(Button2_PIN))
    return(2);
  if(!ReadPin(Button3_PIN))
    return(3);

  return(0);
}

/* First draw points followed by lines to create 3d ball kind of effect */
void GPU_Ball_Stencil(void)
{
  int16_t xball = (HSIZE/2),yball = 120,rball = (HSIZE/8)+3;
  int16_t numpoints = 6,numlines = 8,i,asize,aradius,gridsize = 10;
  int32_t asmooth,loopflag = 1,dispr = (HSIZE - 1),displ = 1,dispa = 1,dispb = (VSIZE - 1),xflag = 1,yflag = 1;
  uint32_t iteration = 0;
  uint8_t red = 255, blue = 255, green = 255;
  uint32_t RGB = 0;  
  int64_t gradientIteration = 0;
  uint32_t touchValue = 0;

  
  
  
  dispr -= ((dispr - displ)%gridsize);
  dispb -= ((dispb - dispa)%gridsize); 
  
  // write the play sound 
  wr16(REG_SOUND + RAM_REG,0x50);
  wr8(REG_VOL_SOUND + RAM_REG,100);        
  
  touchValue = rd32(REG_TOUCH_DIRECT_XY + RAM_REG);// MO: Read for any new touch tag inputs
  
  while(touchValue & 0x80000000)
  {    
    touchValue = rd32(REG_TOUCH_DIRECT_XY + RAM_REG);// MO: Read for any new touch tag inputs
    if(((xball + rball + 1) >= dispr) || ((xball - rball - 1) <= displ))
    {
      xflag ^= 1;
     //wr8(REG_PLAY + RAM_REG,1);
    }
    if(((yball + rball + 2) >= dispb) || ((yball - rball - 2) <= dispa))
    {
      yflag ^= 1;
      //wr8(REG_PLAY+ RAM_REG,1);
    }
    if(xflag)
    {
      xball += 1;
    }
    else
    {
      xball -= 1;
    }
    if(yflag)
    {
      yball += 2 ;
    }
    else
    {
      yball -= 2;
    }    

    Send_CMD(CMD_DLSTART);
    Send_CMD(CLEAR_COLOR_RGB(red, blue, green));    
    Send_CMD(CLEAR(1, 1, 1)); // clear screen
               
    Send_CMD(POINT_SIZE((rball)*16) );
    Send_CMD(BEGIN(POINTS));    

    //shadow
    Send_CMD(COLOR_RGB(0, 0, 0) );//shadow
    Send_CMD(COLOR_A(160) );
    Send_CMD(VERTEX2F((xball+16)*16,(yball+8)*16));

    //black border
    Send_CMD(COLOR_A(255) );
    Send_CMD(COLOR_RGB(0, 0, 0) );
    Send_CMD(POINT_SIZE((rball+3)*16) );
    Send_CMD(VERTEX2F((xball)*16,yball*16));

    //red dot
    Send_CMD(POINT_SIZE(rball*16) );
    Send_CMD(COLOR_RGB(255, 0, 0) );
    Send_CMD(VERTEX2F(xball*16,yball*16));

    Send_CMD(END());

    Send_CMD(DISPLAY());

    Send_CMD(CMD_SWAP);  
    UpdateFIFO();                                            // Trigger the CoProcessor to start processing the FIFO    

    MyDelay(16);    
    iteration++;
    Log("%d iteration\n", iteration);
    Log("%d RGB\n", RGB);
    Log("%d Red\n", red);
    Log("%d blue\n", blue);
    Log("%d green\n", green);
  }
}

/*float lerp(float t, float a, float b)
{
  return (float)((1 - t) * a + t * b);
}

float smoothlerp(float t, float a, float b)
{

    float lt = 3 * t * t - 2 * t * t * t;

    return lerp(lt, a, b);
}*/

/* void GPU_Ball_Stencil(void)
{
  int16_t xball = (HSIZE/2),yball = 120,rball = (HSIZE/8);
  int16_t numpoints = 6,numlines = 8,i,asize,aradius,gridsize = 20;
  int32_t asmooth,loopflag = 1,dispr = (HSIZE - 10),displ = 10,dispa = 10,dispb = (VSIZE - 10),xflag = 1,yflag = 1;

  dispr -= ((dispr - displ)%gridsize);
  dispb -= ((dispb - dispa)%gridsize);
  // write the play sound 
  wr16(REG_SOUND + RAM_REG,0x50);
  wr8(REG_VOL_SOUND + RAM_REG,100);      
  //wr8(REG_PLAY + RAM_REG,1);  
  while(1)
  {
    if(((xball + rball + 2) >= dispr) || ((xball - rball - 2) <= displ))
    {
      xflag ^= 1;
     wr8(REG_PLAY + RAM_REG,1);
    }
    if(((yball + rball + 8) >= dispb) || ((yball - rball - 8) <= dispa))
    {
      yflag ^= 1;
      wr8(REG_PLAY+ RAM_REG,1);
    }
    if(xflag)
    {
      xball += 2;
    }
    else
    {
      xball -= 2;
    }
    if(yflag)
    {
      yball += 8 ;
    }
    else
    {
      yball -= 8;
    }

    Send_CMD(CMD_DLSTART);
    Send_CMD(CLEAR_COLOR_RGB(128,128,0));    
    Send_CMD(CLEAR(1, 1, 1)); // clear screen
    Send_CMD(STENCIL_OP(INCR,INCR) );
    Send_CMD(COLOR_RGB(0, 0, 0) );
    Send_CMD(COLOR_MASK(0,0,0,0) );//mask all the colors
    Send_CMD(POINT_SIZE(rball*16) );
    Send_CMD(BEGIN(POINTS));
    Send_CMD(VERTEX2F(xball*16,yball*16));
    Send_CMD(STENCIL_OP(INCR,0UL) );
    Send_CMD(STENCIL_FUNC(GEQUAL,1,255));
    // one side points 

    for(i=1;i<=numpoints;i++)
    {
      asize = i*rball*2/(numpoints + 1);
      asmooth = (int16_t)smoothlerp((float)((float)(asize)/(2*(float)rball)),0,2*(float)rball);

      if(asmooth > rball)
      {
        //change the offset to -ve
        int32_t tempsmooth;
        tempsmooth = asmooth - rball;
        aradius = (rball*rball + tempsmooth*tempsmooth)/(2*tempsmooth);
        Send_CMD(POINT_SIZE(aradius*16) );
        Send_CMD(VERTEX2F((xball - aradius + tempsmooth)*16,yball*16));
      }
      else
      {
        int32_t tempsmooth;
        tempsmooth = rball - asmooth;
        aradius = (rball*rball + tempsmooth*tempsmooth)/(2*tempsmooth);
        Send_CMD(POINT_SIZE(aradius*16) );
        Send_CMD(VERTEX2F((xball+ aradius - tempsmooth)*16,yball*16));
      }
    }



    Send_CMD(END());
    Send_CMD(BEGIN(LINES));
    // draw lines - line should be at least radius diameter 
    for(i=1;i<=numlines;i++)
    {
      asize = (i*rball*2/numlines);
      asmooth = (int16_t)smoothlerp((float)((float)(asize)/(2*(float)rball)),0,2*(float)rball);
      Send_CMD(LINE_WIDTH(asmooth * 16));
      Send_CMD(VERTEX2F((xball - rball)*16,(yball - rball )*16));
      Send_CMD(VERTEX2F((xball + rball)*16,(yball - rball )*16));
    }
    Send_CMD(END());

    Send_CMD(COLOR_MASK(1,1,1,1) );//enable all the colors
    Send_CMD(STENCIL_FUNC(ALWAYS,1,255));
    Send_CMD(STENCIL_OP(KEEP,KEEP));
    Send_CMD(COLOR_RGB(255, 255, 255) );
    Send_CMD(POINT_SIZE(rball*16) );
    Send_CMD(BEGIN(POINTS));
    Send_CMD(VERTEX2F((xball - 1)*16,(yball - 1)*16));
    Send_CMD(COLOR_RGB(0, 0, 0) );//shadow
    Send_CMD(COLOR_A(160) );
    Send_CMD(VERTEX2F((xball+16)*16,(yball+8)*16));
    Send_CMD(COLOR_A(255) );
    Send_CMD(COLOR_RGB(255, 255, 255) );
    Send_CMD(VERTEX2F(xball*16,yball*16));
    Send_CMD(COLOR_RGB(255, 0, 0) );
    Send_CMD(STENCIL_FUNC(GEQUAL,1,1));
    Send_CMD(STENCIL_OP(KEEP,KEEP));
    Send_CMD(VERTEX2F(xball*16,yball*16));

    Send_CMD(END());

    Send_CMD(DISPLAY());

    Send_CMD(CMD_SWAP);  
    UpdateFIFO();                                            // Trigger the CoProcessor to start processing the FIFO    

    MyDelay(30);    

  }
}*/
